/*
 * Copyright (c) 2010, CollabNet
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of CollabNet nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL COLLABNET BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.collabnet.labmanagement.api;

import static org.codehaus.jackson.annotate.JsonAutoDetect.Visibility.ANY;
import static org.codehaus.jackson.annotate.JsonAutoDetect.Visibility.NONE;

import org.apache.commons.httpclient.HttpClient;
import org.codehaus.jackson.map.DeserializationConfig.Feature;
import org.codehaus.jackson.map.ObjectMapper;
import org.codehaus.jackson.map.introspect.VisibilityChecker.Std;

import java.io.IOException;
import java.net.URL;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * Represents a cloud in Cubit.
 *
 * @author Kohsuke Kawaguchi
 */
public class Cloud implements Named {
    private String name;
    private int isPublic;   // 0 and 1 used as boolean
    private HostType[] host_types;

    public static Cloud parse(URL cubit, String name, HttpClient httpClient, String username, String password) throws IOException {
        String content = LabManagementConnector.loginAndRetrieve(cubit, "/cubit/Clouds/details/" + name, httpClient, username, password);
        String MARKER = "cloud_json =";
        int start = content.lastIndexOf(MARKER);

        return MAPPER.readValue(content.substring(start + MARKER.length(), content.length()), Cloud.class);
    }

    /**
     * Gets the name of the cloud.
     * This is the token that shows up as NAME in URL like
     * https://mgr.example.com/cubit/Clouds/details/NAME
     */
    public String getName() {
        return name;
    }

    /**
     * Is this cloud publicly accessible?
     *
     * TODO: what does this mean, really? I thought I either have an access to a cloud or I don't,
     * and if I don't have an access I wouldn't be able to see it to begin with.
     */
    public boolean isPublic() {
        return isPublic!=0;
    }

    /**
     * All the host types defined for this cloud.
     */
    public List<HostType> getHostTypes() {
        return Collections.unmodifiableList(Arrays.asList(host_types));
    }

    /**
     * Gets the specific host type by its name, or null.
     */
    public HostType getHostType(String name) {
        for (HostType ht : host_types)
            if (ht.getName().equals(name))
                return ht;
        return null;
    }

    /*package*/ static final ObjectMapper MAPPER = new ObjectMapper();

    static {
        MAPPER.setVisibilityChecker(new Std(NONE, NONE, NONE, NONE, ANY));
        MAPPER.getDeserializationConfig().set(Feature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    }

}
