/*
 * Copyright (c) 2010, CollabNet
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of CollabNet nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL COLLABNET BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.collabnet.labmanagement.api;

import org.apache.commons.codec.binary.Hex;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.io.IOUtils;
import org.dom4j.Document;
import org.dom4j.Element;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.net.URL;
import java.net.URLEncoder;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Logger;

public class LabManagementImpl implements InvocationHandler, HandCodedCubit {
    /**
     * Endpoint to connect to, such as "https://host/cubit_api/"
     */
    private final URL endpoint;
    private final HttpClient client;

    /**
     * The login id of the user to connect as.
     */
    private final String user;

    /**
     * The password. Can be null, but if not specified, some of the operations will not work.
     */
    private final String password;

    /**
     * The cubit api key of the user.
     */
    private final String key;

    /**
     * Proxy that this implementatino backs.
     */
    private final LabManagement proxy;

    public LabManagementImpl(URL endpoint, HttpClient client, String user, String key, String password) {
        this.endpoint = endpoint;
        this.client = client;
        this.user = user;
        this.password = password;
        this.key = key;
        proxy = (LabManagement) Proxy.newProxyInstance(getClass().getClassLoader(), new Class[]{LabManagement.class}, this);
    }

    /**
     * Returns a nicely typed proxy backed by this.
     */
    public LabManagement asProxy() {
        return proxy;
    }

    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        if (args==null) args = EMPTY_ARRAY;

        Class<?> dc = method.getDeclaringClass();
        if (dc==Object.class || dc==HandCodedCubit.class) {
            // self-forwarding
            try {
                return method.invoke(this,args);
            } catch (InvocationTargetException e) {
                throw e.getCause();
            }
        }
        
        // invoke a web service
        Map<String, String> argMap = toMap(method, args);
        if (method.getAnnotation(Signed.class)!=null)
            argMap = signArgs(argMap);

        return Response.fromXml(callRest(
                method.getAnnotation(Version.class).value(),
                method.getAnnotation(CgiMethod.class).value(),
                buildArgString(argMap)));
    }

    /**
     * Retrieve an authn token for the user from the server.
     *
     * @return A new authn token.
     * @throws IOException if a transport or HTTP protocol error occurs.
     */
    private String getToken() throws IOException {
        Map<String, String> args = new HashMap<String, String>();
        args.put("userid", this.user);
        args.put("txt", null);
        String token = callRest(1,"get_auth_token", buildArgString(args));
        return token != null ? token.trim() : null;
    }

    /**
     * Calls a Cubit signed Rest function, using the two-phase authn scheme.
     *
     * @param args Map of args to cubit api call.
     * @return Raw output from the ReST call, which will be XML.
     * @throws IOException if a transport or HTTP protocol error occurs.
     */
    private Map<String,String> signArgs(Map<String, String> args) throws IOException {
        args.put("userid", this.user); //$NON-NLS-1$
        TreeMap<String, String> sortedArgs = new TreeMap<String, String>(args);
        sortedArgs.put("sig", buildSig(sortedArgs, getToken()));
        return sortedArgs;
    }

    /**
     * Builds the signature key from the args, token, and user key.
     *
     * Algorithm is:
     *   <li> Sort arguments by key alphanumerically.
     *   <li> Place args in string of format "arg1=val1 arg2=val2..."
     *   <li> Append token & secret (the key)
     *   <li> Get SHA hash of this string, and return hex version of said hash.
     *
     * @param sortedArgs TreeMap of sorted args to cubit api call.
     * @param token Token returned from getToken.
     * @return Signature key string.
     */
    private String buildSig(TreeMap<String, String> sortedArgs, String token) throws IOException {
        StringBuffer argBuf = new StringBuffer();
        for (Map.Entry<String, String> entry : sortedArgs.entrySet()) {
            if (!"sig".equals(entry.getKey())) { //$NON-NLS-1$
                argBuf.append(entry.getKey()).append('=').append(entry.getValue());
                argBuf.append(' ');
            }
        }
        argBuf.append("token=").append(token).append(' '); //$NON-NLS-1$
        argBuf.append("secret=").append(this.key); //$NON-NLS-1$
        // Recreate python 'repr' operation on string.  Hee!
        argBuf.insert(0, "u'").append("'"); //$NON-NLS-1$ //$NON-NLS-2$

        try {
            LOGGER.fine("Sig src: [" + argBuf.toString() + "]");
            MessageDigest md = MessageDigest.getInstance("SHA-1");
            byte[] hash = md.digest(argBuf.toString().getBytes("utf-8"));
            return new String(Hex.encodeHex(hash,true));
        } catch (NoSuchAlgorithmException e) {
            throw (IOException)new IOException("SHA-1 not available").initCause(e);
        }
    }

    private String buildArgString(Map<String, String> args) throws UnsupportedEncodingException {
        StringBuffer argString = new StringBuffer();
        if (args != null) {
            boolean first = true;
            for (Map.Entry<String, String> entry: args.entrySet()) {
                if (first) {
                    first = false;
                } else {
                    argString.append('&');
                }
                String key = entry.getKey();
                String val = entry.getValue();

                key = URLEncoder.encode(key, "utf-8"); //$NON-NLS-1$
                argString.append(key);
                if (val != null) {
                    val = URLEncoder.encode(val, "utf-8"); //$NON-NLS-1$
                }
                argString.append('=').append(val);
            }
        }
        return argString.toString();
    }
    
    /**
     * Converts argument map to URL-friendly argument string.
     */
    private Map<String,String> toMap(Method method, Object[] args) throws UnsupportedEncodingException {
        Map<String,String> r = new HashMap<String,String>();

        Annotation[][] pa = method.getParameterAnnotations();
        for (int i=0; i<args.length; i++) {
            Param p=null;
            for (Annotation a : pa[i]) {
                if (a.annotationType()==Param.class) {
                    p = (Param)a;
                    break;
                }
            }
            if (p==null)    throw new IllegalStateException("Bug in code generation. Missing @Param");

            String name = p.value();
            Object val = args[i];

            if (val!=null)
                r.put(name,val.toString());
        }
        return r;
    }

    /**
     * Makes a REST call.
     */
    private String callRest(int version, String apiName, String argString) throws IOException {
        StringBuilder url = new StringBuilder(endpoint.toString());
        url.append('/').append(version).append('/').append(apiName);
        if (argString != null) {
            url.append('?').append(argString);
        }

        GetMethod httpget = new GetMethod(url.toString());
        InputStream is=null;
        try {
            int rsp = client.executeMethod(httpget);
            is = httpget.getResponseBodyAsStream();
            String msg = IOUtils.toString(is, httpget.getResponseCharSet());
            if (rsp>=400)
                throw new IOException("Failed to call "+url+" responseCode="+rsp+"\n"+msg);
            return msg;
        } finally {
            IOUtils.closeQuietly(is);
            httpget.releaseConnection();
        }
    }

    public String getHostState(String host) throws IOException {
        Response rsp = proxy.queryAlloc(host, "true", 1, "host", 0);
        Element state = (Element) rsp.getRootElement().selectSingleNode("project/host/labels/label[@name='cubit_state']");
        return state.getText();
    }

    public List<String> getCloudNames() throws IOException {
        Document doc = LabManagementConnector.loginAndParse(endpoint, "/cubit/Clouds/", client, user, password);
        List<Element> list = doc.selectNodes("//DIV[@id='cloud_main']//TABLE[@class='lineitems']/TR/TD[1]/SPAN/A");

        List<String> r = new ArrayList<String>();
        for (Element e : list)
            r.add(e.getTextTrim());
        return r;
    }

    public Cloud getCloud(String name) throws IOException {
        return Cloud.parse(this.endpoint,name,client,user,password);
    }

    private static final Object[] EMPTY_ARRAY = new Object[0];
    private static final Logger LOGGER = Logger.getLogger(LabManagementImpl.class.getName());
}
