/*
 * Copyright (c) 2010, CollabNet
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of CollabNet nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL COLLABNET BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package com.collabnet.labmanagement.api;

import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;

import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;

/**
 * Represents a response from Lab Manager web service.
 *
 * @author Kohsuke Kawaguchi
 */
public class Response {
    private final Document document;

    public Response(Document document) {
        this.document = document;
    }

    /**
     * Parses XML in a string into DOM.
     *
     * Parsing a result into String and then into DOM is not efficient, but so far in Cubit
     * the response XML seems to be rather small to justify this.
     */
    public static Response fromXml(String xml) throws DocumentException {
        return new Response(new SAXReader().read(new StringReader(xml)));
    }

    /**
     * Returns true if the response indicates a success.
     */
    public boolean isOK() {
        return "OK".equals(getRootElement().elementText("status"));
    }

    /**
     * Root element of the response XML.
     */
    public Element getRootElement() {
        return document.getRootElement();
    }

    /**
     * Checks if this response indicates a success, and if not, throw an IOException.
     */
    public void check() throws IOException {
        if (!isOK())
            throw new IOException(getRootElement().asXML());
    }

    /**
     * If this response indicates an error, obtain the error message.
     * 
     */
    public String getErrorMessage() {
        return getRootElement().elementText("error");
    }

    /**
     * For those method calls that just return a single output,
     * obtain that text.
     */
    public String getOutput() {
        return getRootElement().elementText("output");
    }

    /**
     * For those method calls that return multiple outputs,
     * obtain them in the list.
     *
     * @return can be empty but never null.
     */
    public List<String> getOutputs() {
        List<String> r = new ArrayList<String>();
        for (Element e : (List<Element>) getRootElement().elements("output"))
            r.add(e.getText());
        return r;
    }

    @Override
    public String toString() {
        return document.asXML();
    }
}
